#!/usr/bin/env python3

from itertools import chain
from pathlib import Path

from hgen.proto import Protos
from hgen.utils import TAB, cprint, cstr


def _crawl_prototypes(src_path: Path) -> "list[Protos]":
    def crawl_file(src: Path):
        status, res = "green", ""
        try:
            target = Protos(src)
            res = f"(x{len(target)})"
            # return target
            results.append(target)
        except ValueError:
            status, res = "yellow", "none"
        finally:
            print(f"{cstr(status, f'{res} in')} {cstr('blue', src.name)}")

    def checked(results: "list[Protos]"):
        if not len(results):
            raise ValueError(
                cstr(
                    "red",
                    f"no function prototypes found in {src_path}",
                )
            )
        return results

    results: list[Protos] = []
    cprint("magenta", f">> from {src_path}")

    if not src_path.is_dir():
        crawl_file(src_path)
        return checked(results)

    for src in sorted(src_path.glob("*.c")):
        crawl_file(src)

    return checked(results)


# this is the dirtiest loop i've ever laid my hands on.
# but it works.
def _align_protos_indentation(protolist: "list[Protos]"):
    def before_len(proto):
        return len(proto.split(TAB)[0])

    def get_longest_prototype_len():
        result = 0
        for container in protolist:
            for proto in container:
                result = max(result, before_len(proto))

        return result + 4  # tabs

    def in_loop():
        types, funcname_params = proto.split(TAB)
        funcname, param = funcname_params.split("(")
        params = param.split(", ")

        to_pad = longest // 4 - before_len(proto) // 4
        nl_tabs = TAB * (1 + len(types) // 4 + to_pad)

        result = [f"{types}{TAB * to_pad}{funcname}("]
        # print(result)
        i = 0
        while True:
            is_first = True
            while len(params) and len(result[i].replace(TAB, "....") + params[0]) < 79:
                if is_first:
                    is_first = False
                else:
                    result[i] += " "
                result[i] += params.pop(0)
                if len(params):
                    result[i] += ","
            i += 1
            if len(params):
                result.append(nl_tabs)
            else:
                break
        return result

    longest = get_longest_prototype_len()

    for container in protolist:
        # print(container)
        results = []
        for proto in container:
            result = in_loop()
            results.append("\n".join(result))

        container.prototypes = results


def _protolist_to_strlist(protolist: "list[Protos]") -> "list[str]":
    return [repr(s) for s in chain(protolist)]


def get_prototypes(src_path: Path) -> "list[str]":
    protolist = _crawl_prototypes(src_path)
    _align_protos_indentation(protolist)
    return _protolist_to_strlist(protolist)


if __name__ == "__main__":
    try:
        results = get_prototypes(Path("../so_long/src"))
    except ValueError as e:
        print(e, "\n====")
        print(get_prototypes(Path("../so_long/lib/src")))



from pathlib import Path

from hgen.utils import RegexRules, cstr, get_lines_from


def get_flag_span(lines: "list[str]") -> "tuple[int, int]":
    begin = None
    for i, line in enumerate(lines):
        if RegexRules.FLAG_BEGIN.match(line):
            begin = i
        elif begin and any(reg.match(line) for reg in RegexRules.FLAG_END):
            return begin, i
    raise SyntaxError(cstr("red", "function flag header was never found"))


def cut_lines_by_flag_span(lines: "list[str]", span: "tuple[int, int]"):
    return lines[: span[0] + 1], lines[span[1] :]


def try_insert(dest: Path, protos: "list[str]") -> bool:
    lines = get_lines_from(dest)
    try:
        before, after = cut_lines_by_flag_span(lines, get_flag_span(lines))
        dest.write_text("\n".join(before + protos + after))
        return True
        # break
    except SyntaxError:
        return False


def insert_prototypes(dest_path: Path, *, protos: "list[str]") -> None:
    def headererror():
        raise NotImplementedError(
            f"searched {dest_path.absolute()} but could not find "
            "neither header or function definition flags.\n"
            "a header in given directory should contain "
            f"{RegexRules.FLAG_BEGIN}"
        )

    if not dest_path.is_dir():
        oper_ok = try_insert(dest_path, protos)
        if oper_ok:
            return
        else:
            headererror()

    for dest in dest_path.glob("**/*.h"):
        print(dest)
        oper_ok = try_insert(dest, protos)
        if oper_ok:
            break
    else:
        headererror()



from pathlib import Path

from hgen.proto import Protos
from termcolor import cprint


def test_protos():
    generated = repr(Protos(Path("test/given/01.c")))
    expected = Path("test/expected/01.c").read_text()

    print(generated)
    cprint(f"{generated=}", "yellow")
    cprint(f"{expected =}", "green")
    assert expected == generated



#!/usr/bin/env python3

from re import IGNORECASE, compile


def _make_flagged_comment_regex(which):
    COMMENT_BEGIN = r"[/][/*].*"
    return compile(rf"{COMMENT_BEGIN}{which}", IGNORECASE)


class RegexRules:
    """Regex Rules to parse *.c and *.h files."""

    FUNCTION = compile(r"^(?P<type>([\w\*]+\s+)+)(?P<name>[\w\*]+\()")
    FLAG_BEGIN = _make_flagged_comment_regex(r"@(func|functions?)\b")
    FLAG_END = [
        compile("#endif"),
        _make_flagged_comment_regex(r"=+.*=+"),
        _make_flagged_comment_regex(r"@end\b"),
    ]


if __name__ == "__main__":
    print(RegexRules.FUNCTION)
    print(RegexRules.FLAG_BEGIN)



#!/usr/bin/env python3

from argparse import Namespace
from pathlib import Path
from sys import argv, stderr

from hgen.injector import insert_prototypes
from hgen.proto import get_prototypes
from hgen.utils import create_parser, visualize

def update_header_prototypes(args: Namespace):
    prototypes: list[str] = []

    dest_path = args.common / args.includes

    for src in args.sources:
        try:
            src_path = args.common / src
            visualize(dest_path, src_path)
            prototypes.extend(get_prototypes(src_path))
        except:
            pass
    insert_prototypes(dest_path, protos=prototypes)


def main():
    parser = create_parser()
    if len(argv) == 1:
        parser.print_usage(stderr)
        exit()

    args = parser.parse_args()
    update_header_prototypes(args)



from argparse import ArgumentParser
from pathlib import Path

from hgen._version import __version__

from .utils import cstr


def create_parser():
    prog = "hgen"
    parser = ArgumentParser(
        prog=prog,
        usage=f"{prog} [-hv] -I header.h src [src ...] [-c path]",
        description=cstr("green", "HGEN: Header prototype GENerator"),
    )
    parser.add_argument(
        "sources",
        nargs="+",
        type=Path,
        default=Path(".") / "src",
        help=(
            "list of source files to search for prototypes. "
            "either directory or file works"
        ),
        metavar="",
    )
    parser.add_argument(
        "-I",
        "--includes",
        type=Path,
        default=Path(".") / "includes",
        required=True,
        help="Path to header file",
        metavar="",
    )
    parser.add_argument(
        "-c",
        "--common",
        type=Path,
        default=Path(),
        help="Path shared by both src and dest",
        metavar="",
    )
    parser.add_argument(
        "-v", "--version", action="version", version=__version__
    )

    # parser.add_argument(
    #     "-q",
    #     "--quiet",
    #     type=bool,
    #     default=False,
    #     help="Whether to suppress output (defualt:false)",
    #     metavar="",
    # )
    return parser


if __name__ == "__main__":
    from sys import argv

    argv.append("--version")
    parser = create_parser()
    args = parser.parse_args()
    print(args)



#!/usr/bin/env python3

from dataclasses import dataclass

# ===== Constant Values =====


@dataclass(frozen=True)
class ColorCodes:
    RED = "\033[91m"
    GREEN = "\033[92m"
    BLUE = "\033[94m"
    CYAN = "\033[96m"
    WHITE = "\033[97m"
    YELLOW = "\033[93m"
    MAGENTA = "\033[95m"
    GREY = "\033[90m"
    BLACK = "\033[90m"
    DEFAULT = "\033[99m"
    END = "\033[0m"


# ===== Functions =====


def cstr(color: str, string: str):
    return f"{getattr(ColorCodes, color.upper())}{string}{ColorCodes.END}"


def cprint(color: str, string: str):
    print(cstr(color, string))



#!/usr/bin/env python3

from pathlib import Path

from hgen.utils.colors import cstr

TAB = "\t"


def get_lines_from(file: Path) -> "list[str]":
    return file.read_text().splitlines(keepends=False)


def visualize(dest_path: Path, src_path: Path):
    cols = {"d": "", "s": ""}
    for n, path in (("d", dest_path), ("s", src_path)):
        p = path.resolve()
        cols[n] += (
            cstr("cyan", f"{p.parent.parent.name}/")
            + cstr("yellow", f"{p.parent.name}/")
            + cstr("red", p.name)
        )
    print(f"{cols['s']} -> {cols['d']}")



#!/usr/bin/env python3

import re
from dataclasses import dataclass, field
from functools import cached_property
from pathlib import Path

from hgen.utils import RegexRules, cstr, get_lines_from
from termcolor import cprint

_NAME_HEADER = "/*\n** < {filename} > */"
_COLOR_HEADER = "magenta"
_COLOR_TYPE = "red"
_COLOR_VAR = "blue"


@dataclass
class Protos:
    """container for c function prototypes defined inside given source code."""

    file: Path
    prototypes: "list[str]" = field(default_factory=list)

    def __post_init__(self):
        self.load_func_protos_from_file()
        if not len(self.prototypes):
            raise ValueError(
                f"{cstr('yellow', self.file.name)} "
                f"{cstr('red',f'never had any BSD-style C function prototypes')}"
            )

    def is_line_func(self, line: str) -> bool:
        return (
            not any(line.startswith(char) for char in "#{")
            and not any(banned in line for banned in ["static", "main"])
            and RegexRules.FUNCTION.match(line) is not None
        )

    def load_func_protos_from_file(self):
        lines = iter(get_lines_from(self.file))
        for line in lines:
            if not self.is_line_func(line):
                continue
            if line.endswith("{"):
                collected = line.replace("{", "").strip().lstrip()
            else:
                collected = line
                for inparam in lines:
                    if inparam == "{":
                        break
                    elif "{" in inparam:
                        collected += inparam.replace("{", "").strip()
                        break
                    else:
                        collected += inparam.lstrip("\t")
            for inside_code in lines:
                if inside_code == "}":
                    break

            self.prototypes.append(f"{collected};")

    def __len__(self) -> int:
        return len(self.prototypes)

    def __repr__(self) -> str:
        return "\n".join([self.header, "", *self.prototypes])

    def __str__(self) -> str:
        return "\n".join(
            [cstr(_COLOR_HEADER, self.header)]
            + [p for p in self.prototypes]
            # + [Protos.colored_prototype(p) for p in self.prototypes]
            + [""]
        )

    def __getitem__(self, item):
        return self.prototypes[item]

    @cached_property
    def header(self):
        return _NAME_HEADER.format(filename=self.file.name)

    @staticmethod
    def colored_prototype(prototype: str):
        type, name, paramstr = re.split(r"\t+|\(", prototype[:-2])
        result = f'{cstr(_COLOR_TYPE, type)} {cstr("magenta", name)}('
        params = iter(paramstr.split())
        for param in params:
            result += cstr(_COLOR_TYPE, param)
            ptr, name = re.split(r"(^\**)", next(params))[1:]
            result += f" {ptr}{cstr(_COLOR_VAR, name)} "
        return result[:-1] + ");"
